#!/bin/sh

#
# Khoros: $Id$
#
# $Log$
#

# Copyright (C) 1993, 1994, 1995, Khoral Research, Inc., ("KRI").
# All rights reserved.  See $BOOTSTRAP/repos/license/License or run klicense.


#************************************************************
#
#  Routine Name: kmpirun - Script that starts up MPI processes
#
#       Purpose: Starts up MPI processes given a process file
#	     	 that describes the processes to be created
#	         and a connection file that specifies how the
#                processes are to communicate with each other.
#                
#                It assumes that you've set the MPIR_HOME to
#                the correct value.
#
#         Input: process_file - list of processes and application
#                               numbers
#		 connection_file - list of MPI intercommunicator
#                                  connections between applications
#
#    Written By: Nathan Doss
#          Date: 11/7/95
# Modifications: 
#
#************************************************************/


######################
# Where is MPI stuff?# <--------- Set MPIR_HOME.
######################
MPIR_HOME=/usr/local/mpi
if [ "$MPIR_HOME" = "#""MPIR_HOME""#" ] ; then
    MPIR_HOME=`pwd`/..
fi
MPIRUN_HOME=$MPIR_HOME/bin

# Print out usage instructions
if [ $# -ne 2 ] ; then
    echo "Usage: $0 <process_file> <connection_file>"
cat << EOF

The process file should look something like:
    machine1  /path/to/exec1  1
    machine2  /path/to/exec1  1
    machine3  /path/to/exec3  2
    machine4  /path/to/exec3  2
    machine5  /path/to/exec5  3
    
The third field is the "application number" assigned to the
group of a single khoros application.

The connection file should look like:

    1 2
    2 5

The numbers are the connections (input_group  output_group).  

EOF
    exit
fi

# Get a name for the pg file to be passed to mpirun
pgfilename=pg.$$

# Transform the process file to a p4pg file
pgfile=`cat $1  | \
        perl -pe 's/\s+\d+//g; s/^(\S+)\s+/$1 1 /g'`

# Write out the pgfile
echo "local 0" >  $pgfilename
echo "$pgfile" >> $pgfilename

# Execute the program
$MPIRUN_HOME/mpirun -p4pg $pgfilename kmpihost $1 $2

# Remove the pgfile
rm $pgfilename

